/*********************************************************************
 *
 * Iteratively extend the permutation which trying to minimise the
 * total search space.
 *
 * start at startdepth, search for a good insert, if none found then
 * go deeper, if best_trials > maxtrials then stop looking.
 * Otherwise, do the full calculation when depth = maxdepth
 *
 ********************************************************************/

#include "prog.h"

/* Look for the best element to put in pi[depth]
 * Set global var trials to the result found.
 * Update *total with no. of trials used.
 * Try insertions in sequence unless budget runs out.
 * Don't allow any test to take more than limit trials.
 * If trials >= limit on return, then no good elt was found.
 */
double
find_good_depth_elt (mindepth, depth, limit, total)
     int mindepth, depth;
     double limit;
     double *total;
{
  int x, y;			/* elements to move */
  int better;
  double result, best_trials, best_result;

  inserts_done = 0;
  x = depth;
  /* The first one within limit trials is better than this: */
  best_trials = limit;
  best_result = limit * 2;
  /* The first "insert" we try is the current permutation */
  for (y = depth; y <= SA; y++) {
    /* try the insert and compare with best_trials */
    insert (x, y);		/* insert x in y's place and shift up/down */
    setup (depth);
    if (BETTER == 0) {
      result = calculate (mindepth, depth, best_trials);
    } else {
      result = calculate (mindepth, depth, best_trials * 2);
    }
    budget = budget - trials;
    *total += trials;
    if (BETTER == 0) {
      better = ((trials < best_trials)
		|| ((trials == best_trials) && (result < best_result)));
    } else {
      better = (((result < best_result) && (trials < (best_trials * 2)))
		|| ((result == best_result) && (trials < best_trials)));
    }
    if (better) {
      best_trials = trials;
      best_result = result;
      inserts_done++;
    } else {
      /* this insert is no good, so reverse it */
      insert (y, x);
    }				/* fi */
    if (budget <= 0)
      break;
  }				/* od */
  /* Search done: restore setup and return best result found: */
  setup (depth);
  trials = best_trials;
  return (best_result);
}				/* End of find_good_depth_elt */

/* Look for the best insertion (where at least one elt is within depth)
 * whose trials result at current depth is smaller than best_trials.
 * Update *total with no. of trials used.
 * Set global var trials to the result found.
 * try random insertions until budget runs out, or "maxgoes" goes have been tried.
 */
double
find_good_insert (mindepth, depth, best_trials, best_result, total)
     int mindepth, depth;
     double best_trials;
     double best_result;
     double *total;
{
  int x1, y1;			/* elements to move */
  int x2, y2;			/* elements to move */
  double result;
  int go;
  double sample;
  int better;

  /* Keep gcc -Wall happy that we don't use before defining these */
  x2 = 0;
  y2 = 0;
  inserts_done = 0;
  go = 0;
  do {
    do {
      x1 = rnd (SA - mindepth + 1) + mindepth;
      y1 = rnd (SA - mindepth + 1) + mindepth;
    } while (((y1 > depth) && (x1 > depth)) || x1 == y1);
    if (TWO_INSERTS) {
      do {
	x2 = rnd (SA - mindepth + 1) + mindepth;
	y2 = rnd (SA - mindepth + 1) + mindepth;
      } while (((y2 > depth) && (x2 > depth)) || x2 == y2);
    }
    /* try the insert and compare with best_trials */
    insert (x1, y1);		/* insert x1 in y1's place and shift up/down */
    if (TWO_INSERTS) {
      insert (x2, y2);		/* insert x2 in y2's place and shift up/down */
    }
    setup (depth);
    if (BETTER == 0) {
      sample = best_trials;
    } else {
      sample = best_trials * 2;
    }
    result = calculate (mindepth, depth, sample);
    budget = budget - trials;
    *total += trials;
    if (BUSH) {
      if ((result == 0) && (trials < sample)) {
	/* have proved that this branch contributes nothing */
	return (0);
      }
    }
    if (BETTER == 0) {
      better = ((trials < best_trials)
		|| ((trials == best_trials) && (result < best_result)));
    } else {
      better = (((result < best_result) && (trials < (best_trials * 2)))
		|| ((result == best_result) && (trials < best_trials)));
    }
    if (better) {
      best_trials = trials;
      best_result = result;
      inserts_done++;
    } else {
      /* this insert is no good, so reverse it */
      if (TWO_INSERTS) {
	insert (y2, x2);
      }
      insert (y1, x1);
    }				/* fi */
    go++;
    if (printinterval && ((go % printinterval) == 1)) {
      /* print statistics every printinterval goes: */
      (void) printf ("%8.0f,%8.0f    ", best_trials, best_result);
      (void) printf ("depth=%3i, goes=%4i, budget=%6.0f\n", depth, go, budget);
    }
  } while ((budget > 0) && (go <= maxgoes));	/* od */
  /* Search done: restore setup and return best results found: */
  setup (depth);
  trials = best_trials;
  return (best_result);
}				/* End of find_good_insert */

void
do_pre_analysis ()
{
  int startdepth, maxdepth;
  int mindepth, depth;
  double best_trials;
  double best_result, result;
  int retries, i;

  randompi ();
  setup (SA);
  for (;;) {
    startdepth = 1;
    mindepth = 1;
    maxdepth = SA - 1;
    trials = 0;
    best_trials = 0;
    best_result = 0;
    /* start at startdepth, search for a good insert, if none found then
     * go deeper, if best_trials > maxtrials then stop looking.
     * Otherwise, do the full calculation when depth = maxdepth
     */

    for (depth = startdepth; depth <= maxdepth; depth++) {
      /* Find a good elt to put in depth position, if none found then
       * increase mindepth and try again.
       */
    pre_analysis_redo:		/* redo with new mindepth */
      budget = maxtrials;
      best_result = find_good_depth_elt (mindepth, depth, maxtrials / 10, &total_trials);
      best_trials = trials;
      if (best_trials >= (maxtrials / 10)) {
	(void) printf ("Best next elt exceeded %.0f trials at depth %i, mindepth %i\n",
		       maxtrials / 10, depth, mindepth);
	/* adjust mindepth */
	mindepth = depth;
	/* set up first part of psi to map to dummy element (0): */
	for (i = 1; i < mindepth; i++) {
	  psi[i] = 0;
	}
	goto pre_analysis_redo;
      }
      budget = maxtrials;
      printinfo (mindepth, depth, best_trials, best_result);
      /* if depth == mindepth then this is the best we can do at this depth: */
      if (depth > mindepth) {
	/* search for a good insert at this depth;
	 * stay at this depth as long as a insert is found
	 * for up to MAXRETRIES goes
	 */
	for (retries = 1; retries <= MAXRETRIES; retries++) {
	  budget = maxtrials - best_trials;
	  best_result = find_good_insert (mindepth, depth, best_trials, best_result, &total_trials);
	  best_trials = trials;
	  printinfo (mindepth, depth, best_trials, best_result);
	  if (inserts_done == 0)
	    break;
	}			/* next retry */
      }				/* fi (depth > mindepth) */
      if (mindepth == 1) {
	max_depth = depth;
	max_depth_result = best_result;
      }
    }				/* next depth */
    /* now have best perm at depth = maxdepth
     * so try a full calculation:
     */
    if (BUSH) {			/* go on to next method */
      setup (SA);
      break;
    }
    (void) printf ("\n--- Now doing trial calculation with this perm ---\n");
    setup (SA);
    result = calculate (1, SA, total_trials);	/* large budget limit here */
    if (trials <= total_trials) {	/* this perm succeeded */
      (void) printf ("maps = %12.0f,\t", result);
      (void) printf ("in %12.0f trials\n", trials);
      printinfo (1, SA, trials, result);
      break;			/* terminate PRE_ANALYSIS loop */
    } else {
      (void) printf ("Trial calculation failed - increasing budget.\n");
      maxtrials = maxtrials * 2;
      maxgoes = maxgoes * 2;
      printinterval = printinterval * 2;
    }
  }				/* End pre-analysis loop */
}				/* End of do_pre_analyse () */
